

#include <stddef.h>
#include <pkgconf/hal.h>

#ifdef CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS

#define CYGARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <cyg/hal/mpc555_regs.h>

#include <cyg/hal/hal_stub.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_intr.h>

#ifdef CYGDBG_HAL_DEBUG_GDB_THREAD_SUPPORT
#include <cyg/hal/dbg-threads-api.h>    // dbg_currthread_id
#endif

/* The saved flating point registers */
double gdb_floating_point_registers[32];

/* Given a trap value TRAP, return the corresponding signal. */

int __computeSignal (unsigned int trap_number)
{
    switch (trap_number)
    {
    case CYGNUM_HAL_VECTOR_MACHINE_CHECK:
        /* Machine check */
    case CYGNUM_HAL_VECTOR_RESERVED_3:
    case CYGNUM_HAL_VECTOR_RESERVED_4:
        return SIGILL;
      
    case CYGNUM_HAL_VECTOR_ALIGNMENT:
        /* Data access */
        return SIGBUS;
                    
    case CYGNUM_HAL_VECTOR_INTERRUPT:
        /* External interrupt */
      return SIGINT;

    case CYGNUM_HAL_VECTOR_TRACE:
        /* Instruction trace */
        return SIGTRAP;
      
    case CYGNUM_HAL_VECTOR_PROGRAM:
        // The register PS contains the value of SRR1 at the time of
        // exception entry. Bits 11-15 contain information about the
        // cause of the exception. Bits 16-31 the PS (MSR) state.
        switch ((get_register (PS) >> 17) & 0xf){
        case 1:                         /* trap */
            return SIGTRAP;
        case 2:                         /* privileged instruction */
        case 4:                         /* illegal instruction */
            return SIGILL;
        case 8:                         /* floating point */
            return SIGFPE;
        default:                        /* should never happen! */
            return SIGTERM;
        }            

    case CYGNUM_HAL_VECTOR_RESERVED_A:
    case CYGNUM_HAL_VECTOR_RESERVED_B:
        return SIGILL;

    case CYGNUM_HAL_VECTOR_FP_UNAVAILABLE:
        /* FPU disabled */
    case CYGNUM_HAL_VECTOR_FP_ASSIST:
        /* FPU assist */
        return SIGFPE;

    case CYGNUM_HAL_VECTOR_DECREMENTER:
        /* Decrementer alarm */
        return SIGALRM;

    case CYGNUM_HAL_VECTOR_SYSTEM_CALL:
        /* System call */
        return SIGSYS;

    case CYGNUM_HAL_VECTOR_SW_EMUL:
        /* A SW_EMUL is generated instead of PROGRAM for illegal
           instructions. */
        return SIGILL;

    case CYGNUM_HAL_VECTOR_DATA_BP:
    case CYGNUM_HAL_VECTOR_INSTRUCTION_BP:
    case CYGNUM_HAL_VECTOR_PERIPHERAL_BP:
    case CYGNUM_HAL_VECTOR_NMI:
        /* Developer port debugging exceptions. */
        return SIGTRAP;

    case CYGNUM_HAL_VECTOR_ITLB_ERROR:
        /* Invalid instruction access. */
        return SIGBUS;

    case CYGNUM_HAL_VECTOR_DTLB_ERROR:
        /* Invalid data access. */
        return SIGSEGV;
        
    default:
        return SIGTERM;
    }
}


/* Return the trap number corresponding to the last-taken trap. */

int __get_trap_number (void)
{
    // The vector is not not part of the GDB register set so get it
    // directly from the save context.
    return _hal_registers->vector >> 8;
}

/* Set the currently-saved pc register value to PC. This also updates NPC
   as needed. */

void set_pc (target_register_t pc)
{
    put_register (PC, pc);
}


/*----------------------------------------------------------------------
 * Single-step support
 */

/* Set things up so that the next user resume will execute one instruction.
   This may be done by setting breakpoints or setting a single step flag
   in the saved user registers, for example. */

static target_register_t irq_state = 0;

void __single_step (void)
{
    target_register_t msr = get_register (PS);

    // Set single-step flag in the exception context.
    msr |= (MSR_SE | MSR_BE);
    // Disable interrupts.
    irq_state = msr & MSR_EE;
    msr &= ~MSR_EE;

    put_register (PS, msr);
}

/* Clear the single-step state. */

void __clear_single_step (void)
{
    target_register_t msr = get_register (PS);

    // Clear single-step flag in the exception context.
    msr &= ~(MSR_SE | MSR_BE);
    // Restore interrupt state.
    // FIXME: Should check whether the executed instruction changed the
    // interrupt state - or single-stepping a MSR changing instruction
    // may result in a wrong EE. Not a very likely scenario though.
    msr |= irq_state;

    // This function is called much more than its counterpart __single_step.
    // Only re-enable interrupts if they where disabled during the previous
    // cal to __single_step. Otherwise, this function only makes "extra sure"
    // that no trace or branche exception will happen.
    irq_state = 0;      

    put_register (PS, msr);
}


void __install_breakpoints (void)
{
    /* NOP since single-step HW exceptions are used instead of
       breakpoints. */
}

void __clear_breakpoints (void)
{
}


/* If the breakpoint we hit is in the breakpoint() instruction, return a
   non-zero value. */

int
__is_breakpoint_function ()
{
    return get_register (PC) == (target_register_t)&CYG_LABEL_NAME(breakinst);
}


/* Skip the current instruction.  Since this is only called by the
   stub when the PC points to a breakpoint or trap instruction,
   we can safely just skip 4. */

void __skipinst (void)
{
    put_register (PC, get_register (PC) + 4);
}

/* Floating point support */
void get_register_as_bytes(regnames_t which, char * bytes)
{
  if((which >= F0) && (which <= F31))
  {
    char * p = (char *)&gdb_floating_point_registers[which - F0];
    int i;
    for(i = 0; i < 8; i++)
    {
      bytes[i] = *p;
      p++;
    }
  }
  else
  { /* place rubbisch in it */
    char p = 0xaa;
    int i;
    for(i = 0; i < 8; i++)
    {
      bytes[i] = p;
      p = ~p;
    }
  }
}

void put_register_as_bytes(regnames_t which, char * bytes)
{
  if((which >= F0) && (which <= F31))
  {
    char * p = (char *)&gdb_floating_point_registers[which - F0];
    int i;
    for(i = 0; i < 8; i++)
    {
      *p = bytes[i];
      p++;
    }
  }
}

double get_floating_point_register(regnames_t which)
{
  if((which >= F0) && (which <= F31))
    return gdb_floating_point_registers[which - F0];

  else
    return 0;
}

void put_floating_point_register(regnames_t which, double value)
{
  if((which >= F0) && (which <= F31))
    gdb_floating_point_registers[which - F0] = value;
}

void hal_mpc555_get_gdb_regs(void * gdbRegisters, HAL_SavedRegisters * ecosRegisters)
{
  regnames_t reg;
  CYG_ADDRWORD * regval = (CYG_ADDRWORD *)gdbRegisters;

#ifdef CYGDBG_HAL_COMMON_INTERRUPTS_SAVE_MINIMUM_CONTEXT
  for(reg = R0; reg <= R16; reg++)
    regval[reg] = ecosRegisters->d[reg - R0];
  for(reg = F0; reg <= F14; reg++)
    put_floating_point_register(reg, ecosRegisters->f[reg - F0]);
#else
  for(reg = R0; reg <= R31; reg++)
    regval[reg] = ecosRegisters->d[reg - R0];
  for(reg = F0; reg <= F31; reg++)
    put_floating_point_register(reg, ecosRegisters->f[reg - F0]);
#endif  

  regval[PC]  = ecosRegisters->pc;
  regval[PS]  = ecosRegisters->msr;
  regval[CND] = ecosRegisters->cr;
  regval[LR]  = ecosRegisters->lr;
  regval[CNT] = ecosRegisters->ctr;
  regval[XER] = ecosRegisters->xer;
}

void hal_mpc555_set_gdb_regs(HAL_SavedRegisters * ecosRegisters, void * gdbRegisters)
{
  regnames_t reg;
  CYG_ADDRWORD * regval = (CYG_ADDRWORD *)gdbRegisters;
#ifdef CYGDBG_HAL_COMMON_INTERRUPTS_SAVE_MINIMUM_CONTEXT
  for(reg = R0; reg <= R16; reg++)
    ecosRegisters->d[reg - R0] = regval[reg];
  for(reg = F0; reg <= F14; reg++)
    ecosRegisters->f[reg - F0] = get_floating_point_register(reg);
#else
  for(reg = R0; reg <= R31; reg++)
    ecosRegisters->d[reg - R0] = regval[reg];
  for(reg = F0; reg <= F31; reg++)
    ecosRegisters->f[reg - F0] = get_floating_point_register(reg);
#endif  

  ecosRegisters->pc  = regval[PC];
  ecosRegisters->msr = regval[PS];
  ecosRegisters->cr  = regval[CND];
  ecosRegisters->lr  = regval[LR];
  ecosRegisters->ctr = regval[CNT];
  ecosRegisters->xer = regval[XER];
}

#endif // CYGDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
