#ifndef CMSEVENT_CXX
#define CMSEVENT_CXX

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "cyg/io/cmsevent.h"
#include <cyg/kal/kdefs.h>

//-----------------------
// Some local definitions
//-----------------------
#define CONTROLLED_EVENT_SCS     0xe0
#define CONTROLLED_EVENT_CCS     0xe0
#define CONTROLLED_EVENT_RS      0x10
#define CONTROLLED_EVENT_AS      0x10
#define CONTROLLED_EVENT_R       0x01
#define CONTROLLED_EVENT_ENABLED 0x10
#define CONTROLLED_EVENT_FAILURE 0x01
#define SCS_NOTIFY               0x00
#define SCS_SET_CONTROL_STATE    0x20
#define CCS_SET_CONTROL_STATE    0x20

//----------------------
// Class implementations
//----------------------
// CMS_Event
// Constructor
CMS_Event::CMS_Event(CMS_Event::CMS_EventClass a_class,
                     unsigned char a_length)
    : m_class(a_class),
      m_length(a_length)
{
}

// Provide an (empty) virtual destructor
CMS_Event::~CMS_Event()
{
}

// CMS_ClientEvent
// Constructor
CMS_ClientEvent::CMS_ClientEvent(CMS_Event::CMS_EventClass a_class,
                                 unsigned int a_inhibittime,
                                 unsigned char a_priority,
                                 const char * a_name,
                                 CAN_msgid a_txcobid, CAN_msgid a_rxcobid,
                                 can_io_handle_t a_handle,
                                 unsigned char a_length,
                                 bool a_numeric)
    : CMS_Event(a_class, a_length),
      CMS_ClientObject(a_inhibittime, a_priority, a_name, a_numeric),
      m_waiting(false)
{
  switch(a_class)
  {
    case Uncontrolled:
      m_rxcob = new RxCOB(a_rxcobid, a_handle, this);
      break;
    case Controlled:
      m_rxcob = new RxCOB(a_rxcobid, a_handle, this);
      m_txcob = new TxCOB(a_txcobid, a_handle, this, a_inhibittime);
      break;
    case Stored:
    default:
      m_rxcob = new RTR_ClientCOB(a_rxcobid, a_handle, this, a_inhibittime);
      break;
  }
}

// Provide an (empty) virtual destructor
CMS_ClientEvent::~CMS_ClientEvent()
{
}

// dataIndication
void CMS_ClientEvent::dataIndication(unsigned char a_data[8],
                                     unsigned char a_length,
                                     CommunicationObject * /*a_cob*/)
{
  switch(m_class)
  {
    case Uncontrolled:
      _ASSERT(a_length == m_length, "CMS_ClientEvent invalid data");
      eventDataIndication(a_data);
      break;
    case Controlled:
      {
        if((a_data[0] & SCS_SET_CONTROL_STATE) == SCS_SET_CONTROL_STATE)
        {
          _ASSERT(a_length == 1, "CMS_ClientEvent invalid data");
          memcpy(&m_data[0], a_data, a_length);
          m_condition.signal();
        }
        else
        {
          _ASSERT(a_length == m_length + 1, "CMS_ClientEvent invalid data");
          eventDataIndication(&a_data[1]);
        }
      }
      break;
    case Stored:
      if(m_waiting)
      {
        m_waiting = false;
        memcpy(&m_data[0], a_data, a_length);
        m_condition.signal();
      }
      else
        eventDataIndication(&a_data[0]);

      break;
    default:
      break;
  }
}

// CMS_ServerEvent
// Constructor
CMS_ServerEvent::CMS_ServerEvent(CMS_Event::CMS_EventClass a_class,
                                 unsigned int a_inhibittime,
                                 unsigned char a_priority,
                                 const char * a_name,
                                 CAN_msgid a_txcobid, CAN_msgid a_rxcobid,
                                 can_io_handle_t a_handle,
                                 unsigned char a_length,
                                 bool a_numeric)
    : CMS_Event(a_class, a_length),
      CMS_ServerObject(a_inhibittime, a_priority, a_name, a_numeric)
{
  switch(a_class)
  {
    case Uncontrolled:
      m_txcob = new TxCOB(a_txcobid, a_handle, this, a_inhibittime);
      break;
    case Controlled:
      m_rxcob = new RxCOB(a_rxcobid, a_handle, this);
      m_txcob = new TxCOB(a_txcobid, a_handle, this, a_inhibittime);
      break;
    case Stored:
    default:
      m_txcob = new RTR_ServerCOB(a_txcobid, a_handle, this, a_inhibittime);
  }
}

// Provide an (empty) virtual destructor
CMS_ServerEvent::~CMS_ServerEvent()
{
}

// dataIndication
void CMS_ServerEvent::dataIndication(unsigned char a_data[8],
                                     unsigned char a_length,
                                     CommunicationObject * a_cob)
{
  _ASSERT(m_class == Controlled, "CMS_ServerEvent invalid data");
  _ASSERT(a_length == 1,         "CMS_ServerEvent invalid data");

  indication(&a_data[0]);
}

// request
void CMS_ServerEvent::request(unsigned char * a_data)
{
  switch(m_class)
  {
    case Uncontrolled:
      ((TxCOB *)m_txcob)->transmit(a_data, m_length);
      break;
    case Controlled:
      {
      unsigned char data[8];
      data[0] = SCS_NOTIFY;
      data[0] &= CONTROLLED_EVENT_SCS; // Set everything else to zero
      memcpy(&data[1], a_data, m_length);

      ((TxCOB *)m_txcob)->transmit(data, m_length + 1);
      }
      break;
    case Stored:
    default:
      ((RTR_ServerCOB *)m_txcob)->transmit(a_data, m_length);
      break;
  }
}

// The controlled server event
// Constructor
ControlledServerEvent::ControlledServerEvent(unsigned int a_inhibittime,
                                             unsigned char a_priority,
                                             const char * a_name,
                                             CAN_msgid a_txcobid, CAN_msgid a_rxcobid,
                                             can_io_handle_t a_handle,
                                             unsigned char a_length,
                                             bool a_numeric)
    : CMS_ServerEvent(CMS_Event::Controlled,
                      a_inhibittime,
                      a_priority,
                      a_name,
                      a_txcobid, a_rxcobid,
                      a_handle, a_length, a_numeric),
      m_controlstate(ControlledServerEvent::Disabled)
{
}

// Provide an (empty) virtual destructor
ControlledServerEvent::~ControlledServerEvent()
{
}

// indication
void ControlledServerEvent::indication(unsigned char * a_data)
{
  if((a_data[0] & CCS_SET_CONTROL_STATE) == CCS_SET_CONTROL_STATE)
  {
    ControlState state = ((a_data[0] & CONTROLLED_EVENT_ENABLED) == CONTROLLED_EVENT_ENABLED ? Enabled : Disabled);
    unsigned char data[8];
    data[0] = SCS_SET_CONTROL_STATE;

    if(!setControlState(state, &data[1]))
      data[0] |= CONTROLLED_EVENT_FAILURE;
    else
      m_controlstate = state;

    if(m_controlstate == Enabled)
      data[0] |= CONTROLLED_EVENT_ENABLED;

    ((TxCOB *)m_txcob)->transmit(data, m_length + 1);
  }
}

// notifyEvent
CMS_Error ControlledServerEvent::notifyEvent(unsigned char * a_data)
{
  if(m_controlstate == Enabled)
  {
    request(a_data);
    return CMS_NOERR;
  }
  else
    return -CMS_INVALID_REQUEST;
}

// The controlled client event
// Constructor
ControlledClientEvent::ControlledClientEvent(unsigned int a_inhibittime,
                                             unsigned char a_priority,
                                             const char * a_name,
                                             CAN_msgid a_txcobid, CAN_msgid a_rxcobid,
                                             can_io_handle_t a_handle,
                                             unsigned char a_length,
                                             bool a_numeric)
    : CMS_ClientEvent(CMS_Event::Controlled,
                      a_inhibittime, a_priority,
                      a_name,
                      a_txcobid, a_rxcobid,
                      a_handle,
                      a_length,
                      a_numeric)
{
}

// Provide an (empty) virtual destructor
ControlledClientEvent::~ControlledClientEvent()
{
}

// setEventControlState
CMS_Error ControlledClientEvent::setEventControlState(bool a_enable,
                                                      unsigned char * a_response,
                                                      unsigned int a_timeout)
{
  KSingleLock sl(&m_condition);
  setEventControlStateRequest(a_enable);
  if(!sl.lock(a_timeout))
    return -CMS_TIMEOUT;

  sl.unlock();
  
  return setEventControlStateConfirmation(a_response);
} 

// setEventControlStateRequest
void ControlledClientEvent::setEventControlStateRequest(bool a_enable)
{
  unsigned char request = CCS_SET_CONTROL_STATE;
  if(a_enable)
    request |= CONTROLLED_EVENT_RS;

  ((TxCOB *)m_txcob)->transmit(&request, 1);
}

// SetEventControlStateConfirmation
CMS_Error ControlledClientEvent::setEventControlStateConfirmation(unsigned char * a_response)
{
  if(m_data[0] & CONTROLLED_EVENT_FAILURE)
  {
    memcpy(a_response, &m_data[1], m_length);
    return -CMS_APPLICATIONERR;
  }

  return CMS_NOERR;
}

// The uncontrolled server event
// Constructor
UncontrolledServerEvent::UncontrolledServerEvent(unsigned int a_inhibittime,
                                                 unsigned char a_priority,
                                                 const char * a_name,
                                                 CAN_msgid a_txcobid,
                                                 can_io_handle_t a_handle,
                                                 unsigned char a_length,
                                                 bool a_numeric)
    : CMS_ServerEvent(CMS_Event::Uncontrolled, 
                      a_inhibittime,
                      a_priority,
                      a_name,
                      a_txcobid, 0,
                      a_handle,
                      a_length,
                      a_numeric)
{
}

// Provide an (empty) virtual destructor
UncontrolledServerEvent::~UncontrolledServerEvent()
{
}

// indication
void UncontrolledServerEvent::indication(unsigned char * /*a_data*/)
{
  // Base class already guarantees that this is never called; You never know
  _ASSERT(1 == 0, "UncontrolledServerEvent invalid message");
}

// notifyEvent
CMS_Error UncontrolledServerEvent::notifyEvent(unsigned char * a_data)
{
  request(a_data);
  return CMS_NOERR;
}

// The uncontrolled client event
// Constructor
UncontrolledClientEvent::UncontrolledClientEvent(unsigned int a_inhibittime,
                                                 unsigned char a_priority,
                                                 const char * a_name,
                                                 CAN_msgid a_rxcobid,
                                                 can_io_handle_t a_handle,
                                                 unsigned char a_length,
                                                 bool a_numeric)
    : CMS_ClientEvent(CMS_Event::Uncontrolled,
                      a_inhibittime,
                      a_priority,
                      a_name,
                      0, a_rxcobid,
                      a_handle,
                      a_length,
                      a_numeric)
{
}

// Provide an (empty) virtual destructor
UncontrolledClientEvent::~UncontrolledClientEvent()
{
}

// The stored server event
// Constructor
StoredServerEvent::StoredServerEvent(unsigned int a_inhibittime,
                                     unsigned char a_priority,
                                     const char * a_name,
                                     CAN_msgid a_txcobid,
                                     can_io_handle_t a_handle,
                                     unsigned char a_length,
                                     bool a_numeric)
    : CMS_ServerEvent(CMS_Event::Stored, a_inhibittime, a_priority,
                      a_name,
                      a_txcobid, 0,
                      a_handle, a_length, a_numeric)
{
}

// Provide an (empty) virtual destructor
StoredServerEvent::~StoredServerEvent()
{
}

// indication
void StoredServerEvent::indication(unsigned char * /*a_data*/)
{
  // Base class already guarantees that this is never called; you never know
  _ASSERT(1 = 0, "StoredServerEvent received invalid message");
}

// StoreEvent
CMS_Error StoredServerEvent::storeEvent(unsigned char * a_data, bool a_immediatenotify)
{
  if(a_immediatenotify)
    request(a_data);
  else
    ((RTR_ServerCOB *)m_txcob)->update(a_data, m_length);

  return CMS_NOERR;
}

// The stored client event
// Constructor
StoredClientEvent::StoredClientEvent(unsigned int a_inhibittime,
                                     unsigned char a_priority,
                                     const char * a_name,
                                     CAN_msgid a_rxcobid,
                                     can_io_handle_t a_handle,
                                     unsigned char a_length,
                                     bool a_numeric)
    : CMS_ClientEvent(CMS_Event::Stored,
                      a_inhibittime, 
                      a_priority,
                      a_name,
                      0, a_rxcobid,
                      a_handle,
                      a_length,
                      a_numeric)
{
}

// Provide an (empty) virtual destructor
StoredClientEvent::~StoredClientEvent()
{
}

// readEvent
CMS_Error StoredClientEvent::readEvent(unsigned char * a_eventdata, unsigned int a_timeout)
{
  KSingleLock sl(&m_condition);
  requestEvent();
  if(!sl.lock(a_timeout))
    return -CMS_TIMEOUT;

  sl.unlock();
  return responseEvent(a_eventdata);
}

// requestEvent
void StoredClientEvent::requestEvent(void)
{
  m_waiting = true;
  ((RTR_ClientCOB *)m_rxcob)->request(m_length);
}

// responseEvent
CMS_Error StoredClientEvent::responseEvent(unsigned char * a_data)
{
  memcpy(a_data, &m_data[0], m_length);
  return CMS_NOERR;
}

#endif
